<?php
/**
 * SPDX-FileCopyrightText: 2024 Sascha Nitsch (grumpydeveloper) https://contentnation.net/@grumpydevelop
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * @author Sascha Nitsch (grumpydeveloper)
 **/
namespace ScriptedBrowser\Commands;

/**
 * calculate given formula
 *
 * @param \ScriptedBrowser\Main $main main instance
 * @param \ScriptedBrowser\Control $control control interface
 * @param array<string,mixed>  $options options
 * @return bool always true
 */
function calc($main, $control, $options)
{
    require_once 'vendor/autoload.php';
    $input = $main->vars($options['parameter']);
    $result = $options['result'];
    try {
        $calc = \Arokettu\ArithmeticParser\Calculator::evaluate($input);
        $main->setVar($result, $calc);
        return true;
    } catch (\Exception) {
        return false;
    }
}

/**
 * inject and execute given javascript
 *
 * @param \ScriptedBrowser\Main $main main instance
 * @param \ScriptedBrowser\Control $control control interface
 * @param array<string,mixed>  $options options
 * @return bool always true
 */
function injectjscript($main, $control, $options)
{
    $script = $main->vars($options['parameter']);
    $jscript = '{const s = document.createElement("script");s.setAttribute("src", "'.$script.'"); s.setAttribute("type", "text/javascript"); document.head.append(s);}';
    $control->executeScript($jscript);
    return true;
}

/**
 * press given key
 *
 * @param \ScriptedBrowser\Main $main
 * @param \ScriptedBrowser\Control $control
 * @param array<string,mixed> $options options
 * @return bool always true
 */
function key($main, $control, $options)
{
    $key = $main->vars($options['parameter']);
    $control->key($key);
    return true;
}

/**
 * do mouse left click
 *
 * @param \ScriptedBrowser\Main $main
 * @unused-param $main
 * @param \ScriptedBrowser\Control $control
 * @param array<string,mixed>  $options options
 * @unused-param $options
 */
function leftclick($main, $control, $options) : bool
{
    $control->click(\ScriptedBrowser\MouseButton::Left);
    return true;
}

function leftpress($main, $control, $options) : bool
{
    $control->mouseDown(\ScriptedBrowser\MouseButton::Left);
    return true;
}

function leftrelease($main, $control, $options) : bool
{
    $control->mouseUp(\ScriptedBrowser\MouseButton::Left);
    return true;
}

/**
 * move mouse to given element
 *
 * @param \ScriptedBrowser\Main $main main instance
 * @param \ScriptedBrowser\Control $control control interface
 * @param array<string,mixed>  $options options
 * @return bool true if mouse move, false on error
 */
function mousemove($main, $control, $options) : bool
{
    $element = $control->selectCSS($main->vars($options['parameter']));
    if ($element === [] || array_key_exists('error', $element)) {
        return false;
    }
    $element = array_values($element)[0];
    $targetPos = $control->getElementRect($element);
    $targetX = $targetPos['x'] + $targetPos['width'] / 2;
    $targetY = $targetPos['y'] + $targetPos['height'] / 2;
    $scroll = $control->getScrollPosition();
    $targetX -= $scroll[0];
    $targetY -= $scroll[1];
    // echo "move $element - $tokens[1] $targetX $targetY\n";
    _mousemoveaction($main, $control, $options, $targetX, $targetY);
    return true;
}

/**
 * move mouse to given target position from variables
 *
 * @param \ScriptedBrowser\Main $main main instance
 * @param \ScriptedBrowser\Control $control control interface
 * @param array<string,mixed>  $options options
 * @return bool true if mouse move, false on error
 */
function mousemovetarget($main, $control, $options) : bool
{
    $targetX = $main->vars($options['x']);
    $targetY = $main->vars($options['y']);
    $scroll = $control->getScrollPosition();
    $targetX -= $scroll[0];
    $targetY -= $scroll[1];
    _mousemoveaction($main, $control, $options, $targetX, $targetY);
    return true;
}

/**
 * move mouse to given target position
 *
 * @param \ScriptedBrowser\Main $main main instance
 * @param \ScriptedBrowser\Control $control control interface
 * @param array<string,mixed>  $options options
 */
function _mousemoveaction($main, $control, $options, $targetX, $targetY)
{
    $mouseOffset = $main->getMouseOffset();
    $mouseX = $main->getMouseX();
    $mouseY = $main->getMouseY();
    $delta = [$targetX - $mouseX, $targetY - $mouseY];
    $framerate = $main->getFrameRate();
    $dist = sqrt($delta[0]*$delta[0] + $delta[1]*$delta[1]);
    if (array_key_exists('duration', $options)) {
        $steps = $options['duration'] * $framerate;
    } else {
        $speed = array_key_exists('mousespeed', $options) ? $options['mousespeed'] : $main->getMouseSpeed();
        $steps = $dist * $main->getFrameRate() / $speed;
    }
    for ($i = 0; $i < $steps; ++$i) {
        $ease = \ScriptedBrowser\bezierBlend($i / $steps);
        $x = $mouseX + $delta[0] * $ease;
        $y = $mouseY + $delta[1] * $ease;
        $control->mousemove(intval($x + $mouseOffset[0]), intval($y + $mouseOffset[1]));
        usleep(1000000/$framerate);
    }
    $main->setMouseX($targetX);
    $main->setMouseY($targetY);
    $control->mousemove($targetX + $mouseOffset[0], $targetY + $mouseOffset[1]);
}

/**
 * move mouse relative
 *
 * @param \ScriptedBrowser\Main $main main instance
 * @unused-param $main
 * @param \ScriptedBrowser\Control $control control interface
 * @param array<string,mixed>  $options options
 * @return bool true if mouse move, false on error
 */
function mousemoverel($main, $control, $options)
{
    $coord = explode(" ", $options['parameter']);
    $x = intval($coord[0], 10);
    $y = intval($coord[1], 10);
    if (array_key_exists('jump', $options)) {
        $control->mousemoveRel($x, $y);
    } else {
        $mouseX = $main->getMouseX();
        $mouseY = $main->getMouseY();
        _mousemoveaction($main, $control, $options, $mouseX + $x, $mouseY + $y);
    }
    return true;
}

/**
 * scroll
 *
 * @param \ScriptedBrowser\Main $main
 * @unused-param $main
 * @param \ScriptedBrowser\Control $control control interface
 * @unused-param $control
 * @param array<string,mixed>  $options options
 * @return bool always true
 */
function scroll($main, $control, $options)
{
    $amount = $options['parameter'];
    $delay = $options['delay'];
    $control->scroll(intval($amount), floatval($delay));
    return true;
}

/**
 * sleep
 *
 * @param \ScriptedBrowser\Main $main
 * @unused-param $main
 * @param \ScriptedBrowser\Control $control control interface
 * @unused-param $control
 * @param array<string,mixed>  $options options
 * @return bool always true
 */
function sleep($main, $control, $options) : bool
{
    usleep(floatval($options['parameter']) * 1000000);
    return true;
}

/**
 * type given text
 *
 * @param \ScriptedBrowser\Main $main
 * @param \ScriptedBrowser\Control $control
 * @param array<string,mixed> $options options
 * @return bool always true
 */
function type($main, $control, $options)
{
    $text = $main->vars($options['parameter']);
    $speed = array_key_exists('typespeed', $options) ? $options['typespeed'] : $main->getTypeSpeed();
    $control->type($speed, $text);
    return true;
}

/**
 * go to url
 *
 * @param \ScriptedBrowser\Main $main main instance
 * @param \ScriptedBrowser\Control $control control interface
 * @param array<string,mixed>  $options options
 * @return bool true if url was changed, false is timeout
 */
function url($main, $control, $options) : bool
{
    $currentURL = $control->getURL();
    $url = $main->vars($options['parameter']);
    if ($currentURL !== $url) {
        $control->setURL($url);
    }
    return waitforurl($main, $control, ['parameter' => $url]);
}

/**
 * wait until given url is loaded
 *
 * @param \ScriptedBrowser\Main $main main instance
 * @param \ScriptedBrowser\Control $control control interface
 * @param array<string,mixed>  $options options
 * @return boolean true if URL changed to target, false is timeout
 */
function waitforurl($main, $control, $options) : bool
{
    $target = $main->vars($options['parameter']);
    $timeout = array_key_exists('timeout', $options) ? $options['timeout'] : 5;
    $pollInterval = array_key_exists('pollInterval', $options) ? $options['pollInterval'] : 100;
    $cancel = hrtime(true) + $timeout * 1000000000;
    do {
        $currentURL = $control->getURL();
        if ($currentURL === $target) {
            return true;
        }
        usleep(1000 * $pollInterval);
    } while (hrtime(true) < $cancel);
    return false;
}
